# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
import types

from math import fabs
from weakref import ref as weakref_ref

from pyomo.common.autoslots import AutoSlots
from pyomo.common.deprecation import deprecation_warning, RenamedClass
from pyomo.common.errors import PyomoException
from pyomo.common.log import is_debug_set
from pyomo.common.numeric_types import native_logical_types, native_types
from pyomo.common.modeling import unique_component_name, NOTSET
from pyomo.common.timing import ConstructionTimer
from pyomo.core import (
    ModelComponentFactory,
    Binary,
    Block,
    Constraint,
    ConstraintList,
    Any,
    LogicalConstraint,
    LogicalConstraintList,
    BooleanValue,
    ScalarBooleanVar,
    ScalarVar,
    value,
)
from pyomo.core.base.component import (
    ActiveComponent,
    ActiveComponentData,
    ComponentData,
)
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.block import BlockData
from pyomo.core.base.misc import apply_indexed_rule
from pyomo.core.base.indexed_component import ActiveIndexedComponent
from pyomo.core.expr.expr_common import ExpressionType

logger = logging.getLogger('pyomo.gdp')

_rule_returned_none_error = """Disjunction '%s': rule returned None.

Disjunction rules must return an iterable containing Disjuncts or
individual expressions, or Disjunction.Skip.  The most common cause of
this error is forgetting to include the "return" statement at the end of
your rule.
"""


class GDP_Error(PyomoException):
    """Exception raised while processing GDP Models"""


class AutoLinkedBinaryVar(ScalarVar):
    """A binary variable implicitly linked to its equivalent Boolean variable.

    Basic operations like setting values and fixing/unfixing this
    variable are also automatically applied to the associated Boolean
    variable.

    As this class is only intended to provide a deprecation path for
    Disjunct.indicator_var, it only supports Scalar instances and does
    not support indexing.
    """

    INTEGER_TOLERANCE = 0.001

    __autoslot_mappers__ = {'_associated_boolean': AutoSlots.weakref_mapper}

    def __init__(self, boolean_var=None):
        super().__init__(domain=Binary)
        self._associated_boolean = weakref_ref(boolean_var)

    def get_associated_boolean(self):
        return self._associated_boolean()

    def set_value(self, val, skip_validation=False, _propagate_value=True):
        super().set_value(val, skip_validation)
        if not _propagate_value:
            return
        # Map the incoming (numeric) value to bool/None
        if val is None:
            bool_val = None
        elif fabs(val - 0.5) < 0.5 - AutoLinkedBinaryVar.INTEGER_TOLERANCE:
            bool_val = None
        else:
            bool_val = bool(int(val + 0.5))
        # (Setting _propagate_value prevents infinite recursion.)
        self.get_associated_boolean().set_value(
            bool_val, skip_validation, _propagate_value=False
        )

    def fix(self, value=NOTSET, skip_validation=False):
        super().fix(value, skip_validation)
        bool_var = self.get_associated_boolean()
        if not bool_var.is_fixed():
            bool_var.fix()

    def unfix(self):
        super().unfix()
        bool_var = self.get_associated_boolean()
        if bool_var.is_fixed():
            bool_var.unfix()


class AutoLinkedBooleanVar(ScalarBooleanVar):
    """A Boolean variable implicitly linked to its equivalent binary variable.

    This class provides a deprecation path for GDP.  Originally,
    Disjunct indicator_var was a binary variable.  This simplified early
    transformations.  However, with the introduction of a proper logical
    expression system, the mathematically correct approach is for the
    Disjunct's indicator_var attribute to be a proper BooleanVar.  As
    part of the transition, indicator_var attributes are instances of
    AutoLinkedBooleanVar, which allow the indicator_var to be used in
    logical expressions, but also implicitly converted (with deprecation
    warning) into their equivalent binary variable.

    Basic operations like setting values and fixing/unfixing this
    variable are also automatically applied to the associated binary
    variable.

    As this class is only intended to provide a deprecation path for
    Disjunct.indicator_var, it only supports Scalar instances and does
    not support indexing.

    """

    def as_numeric(self):
        """Return the binary variable associated with this Boolean variable.

        This method returns the associated binary variable along with a
        deprecation warning about using the Boolean variable in a numeric
        context.

        """
        deprecation_warning(
            "Implicit conversion of the Boolean indicator_var '%s' to a "
            "binary variable is deprecated and will be removed.  "
            "Either express constraints on indicator_var using "
            "LogicalConstraints or work with the associated binary "
            "variable from indicator_var.get_associated_binary()" % (self.name,),
            version='6.0',
        )
        return self.get_associated_binary()

    def as_binary(self):
        return self.as_numeric()

    def set_value(self, val, skip_validation=False, _propagate_value=True):
        # super() does not work as expected for properties; we will call
        # the property setter explicitly.
        super().set_value(val, skip_validation)
        if not _propagate_value:
            return
        # Fetch the current value (so we know it has already been cast
        # to None/bool)
        val = self.value
        if val is not None:
            val = int(val)
        # (Setting _propagate_value prevents infinite recursion.)
        self.get_associated_binary().set_value(
            val, skip_validation, _propagate_value=False
        )

    def fix(self, value=NOTSET, skip_validation=False):
        super().fix(value, skip_validation)
        bin_var = self.get_associated_binary()
        if not bin_var.is_fixed():
            bin_var.fix()

    def unfix(self):
        super().unfix()
        bin_var = self.get_associated_binary()
        if bin_var.is_fixed():
            bin_var.unfix()

    #
    # Duck-type the numeric expression API, but route the conversion to
    # Binary through as_numeric to generate the deprecation warning
    #

    @property
    def bounds(self):
        return self.as_numeric().bounds

    @bounds.setter
    def bounds(self, value):
        self.as_numeric().bounds = value

    @property
    def lb(self):
        return self.as_numeric().lb

    @lb.setter
    def lb(self, value):
        self.as_numeric().lb = value

    @property
    def ub(self):
        return self.as_numeric().ub

    @ub.setter
    def ub(self, value):
        self.as_numeric().ub = value

    def __abs__(self):
        return self.as_numeric().__abs__()

    def __float__(self):
        return self.as_numeric().__float__()

    def __int__(self):
        return self.as_numeric().__int__()

    def __neg__(self):
        return self.as_numeric().__neg__()

    def __bool__(self):
        return self.as_numeric().__bool__()

    def __pos__(self):
        return self.as_numeric().__pos__()

    def get_units(self):
        return self.as_numeric().get_units()

    def has_lb(self):
        return self.as_numeric().has_lb()

    def has_ub(self):
        return self.as_numeric().has_ub()

    def is_binary(self):
        return self.as_numeric().is_binary()

    def is_continuous(self):
        return self.as_numeric().is_continuous()

    def is_integer(self):
        return self.as_numeric().is_integer()

    def polynomial_degree(self):
        return self.as_numeric().polynomial_degree()

    def __le__(self, arg):
        return self.as_numeric().__le__(arg)

    def __lt__(self, arg):
        return self.as_numeric().__lt__(arg)

    def __ge__(self, arg):
        return self.as_numeric().__ge__(arg)

    def __gt__(self, arg):
        return self.as_numeric().__gt__(arg)

    def __eq__(self, arg):
        # If the other operand is a Boolean, then we want to fall back
        # on the "normal" implementation of __eq__ for Boolean values
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return super().__eq__(arg)
        # Otherwise, we will treat this as a binary operand and use the
        # (numeric) relational expression system
        return self.as_numeric().__eq__(arg)

    def __ne__(self, arg):
        # If the other operand is a Boolean, then we want to fall back
        # on the "normal" implementation of __ne__ for Boolean values
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return super().__ne__(arg)
        # Otherwise, we will treat this as a binary operand and use the
        # (numeric) relational expression system
        return self.as_numeric().__ne__(arg)

    def __add__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__add__(arg)

    def __div__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__div__(arg)

    def __mul__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__mul__(arg)

    def __pow__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__pow__(arg)

    def __sub__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__sub__(arg)

    def __truediv__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__truediv__(arg)

    def __iadd__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__iadd__(arg)

    def __idiv__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__idiv__(arg)

    def __imul__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__imul__(arg)

    def __ipow__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__ipow__(arg)

    def __isub__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__isub__(arg)

    def __itruediv__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__itruediv__(arg)

    def __radd__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__radd__(arg)

    def __rdiv__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__rdiv__(arg)

    def __rmul__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__rmul__(arg)

    def __rpow__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__rpow__(arg)

    def __rsub__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__rsub__(arg)

    def __rtruediv__(self, arg):
        if isinstance(arg, BooleanValue) or arg.__class__ in native_logical_types:
            return NotImplemented
        return self.as_numeric().__rtruediv__(arg)

    def setlb(self, arg):
        return self.as_numeric().setlb(arg)

    def setub(self, arg):
        return self.as_numeric().setub(arg)


# The following should eventually be promoted so that all
# IndexedComponents can use it
class _Initializer:
    """A simple function to process an argument to a Component constructor.

    This checks the incoming initializer type and maps it to a static
    identifier so that when constructing indexed Components we can avoid
    a series of isinstance calls.  Eventually this concept should be
    promoted to pyomo.core so that all Components can leverage a
    standardized approach to processing "flexible" arguments (POD data,
    rules, dicts, generators, etc)."""

    value = 0
    deferred_value = 1
    function = 2
    dict_like = 3

    @staticmethod
    def process(arg):
        if type(arg) in native_types:
            return (_Initializer.value, bool(arg))
        elif type(arg) is types.FunctionType:
            return (_Initializer.function, arg)
        elif isinstance(arg, ComponentData):
            return (_Initializer.deferred_value, arg)
        elif hasattr(arg, '__getitem__'):
            return (_Initializer.dict_like, arg)
        else:
            # Hopefully this thing is castable to the type that is desired
            return (_Initializer.deferred_value, arg)


class DisjunctData(BlockData):
    __autoslot_mappers__ = {'_transformation_block': AutoSlots.weakref_mapper}

    _Block_reserved_words = set()

    @property
    def transformation_block(self):
        return (
            None if self._transformation_block is None else self._transformation_block()
        )

    def __init__(self, component):
        BlockData.__init__(self, component)
        with self._declare_reserved_components():
            self.indicator_var = AutoLinkedBooleanVar()
            self.binary_indicator_var = AutoLinkedBinaryVar(self.indicator_var)
        self.indicator_var.associate_binary_var(self.binary_indicator_var)
        # pointer to transformation block if this disjunct has been
        # transformed. None indicates it hasn't been transformed.
        self._transformation_block = None

    def activate(self):
        super(DisjunctData, self).activate()
        self.indicator_var.unfix()

    def deactivate(self):
        super(DisjunctData, self).deactivate()
        self.indicator_var.fix(False)

    def _deactivate_without_fixing_indicator(self):
        super(DisjunctData, self).deactivate()

    def _activate_without_unfixing_indicator(self):
        super(DisjunctData, self).activate()


class _DisjunctData(metaclass=RenamedClass):
    __renamed__new_class__ = DisjunctData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("Disjunctive blocks.")
class Disjunct(Block):
    _ComponentDataClass = DisjunctData

    def __new__(cls, *args, **kwds):
        if cls != Disjunct:
            return super(Disjunct, cls).__new__(cls)
        if args == ():
            return ScalarDisjunct.__new__(ScalarDisjunct)
        else:
            return IndexedDisjunct.__new__(IndexedDisjunct)

    def __init__(self, *args, **kwargs):
        if kwargs.pop('_deep_copying', None):
            # Hack for Python 2.4 compatibility
            # Deep copy will copy all items as necessary, so no need to
            # complete parsing
            return

        kwargs.setdefault('ctype', Disjunct)
        Block.__init__(self, *args, **kwargs)

    # For the time being, this method is not needed.
    #
    # def _deactivate_without_fixing_indicator(self):
    #    # Ideally, this would be a super call from this class.  However,
    #    # doing that would trigger a call to deactivate() on all the
    #    # DisjunctData objects (exactly what we want to avoid!)
    #    #
    #    # For the time being, we will do something bad and directly call
    #    # the base class method from where we would otherwise want to
    #    # call this method.

    def _activate_without_unfixing_indicator(self):
        # Ideally, this would be a super call from this class.  However,
        # doing that would trigger a call to deactivate() on all the
        # DisjunctData objects (exactly what we want to avoid!)
        #
        # For the time being, we will do something bad and directly call
        # the base class method from where we would otherwise want to
        # call this method.
        ActiveComponent.activate(self)
        if self.is_indexed():
            for component_data in self.values():
                component_data._activate_without_unfixing_indicator()


class ScalarDisjunct(DisjunctData, Disjunct):
    def __init__(self, *args, **kwds):
        DisjunctData.__init__(self, self)
        Disjunct.__init__(self, *args, **kwds)
        self._data[None] = self
        self._index = UnindexedComponent_index


class SimpleDisjunct(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarDisjunct
    __renamed__version__ = '6.0'


class IndexedDisjunct(Disjunct):
    #
    # HACK: this should be implemented on ActiveIndexedComponent, but
    # that will take time and a PEP
    #
    @property
    def active(self):
        return any(d.active for d in self._data.values())


DisjunctData._Block_reserved_words = set(dir(Disjunct()))


class DisjunctionData(ActiveComponentData):
    __slots__ = ('disjuncts', 'xor', '_algebraic_constraint', '_transformation_map')
    __autoslot_mappers__ = {'_algebraic_constraint': AutoSlots.weakref_mapper}
    _NoArgument = (0,)

    @property
    def algebraic_constraint(self):
        return (
            None if self._algebraic_constraint is None else self._algebraic_constraint()
        )

    def __init__(self, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - ConstraintData,
        #   - ActiveComponentData
        #   - ComponentData
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._active = True
        self.disjuncts = []
        self.xor = True
        # pointer to XOR (or OR) constraint if this disjunction has been
        # transformed. None if it has not been transformed
        self._algebraic_constraint = None
        # Dictionary to notate information from partial transformations
        self._transformation_map = {}

    def set_value(self, expr):
        for e in expr:
            # The user gave us a proper Disjunct block
            # [ESJ 06/21/2019] This is really an issue with the reclassifier,
            # but in the case where you are iteratively adding to an
            # IndexedDisjunct indexed by Any which has already been transformed,
            # the new Disjuncts are Blocks already. This catches them for who
            # they are anyway.
            if hasattr(e, 'is_component_type') and e.is_component_type():
                if e.ctype == Disjunct and not e.is_indexed():
                    self.disjuncts.append(e)
                    continue
                e_iter = [e]
            elif hasattr(e, '__iter__'):
                e_iter = e
            else:
                e_iter = [e]
            # The user was lazy and gave us a single constraint
            # expression or an iterable of expressions
            expressions = []
            propositions = []
            for _tmp_e in e_iter:
                try:
                    if _tmp_e.is_expression_type(ExpressionType.RELATIONAL):
                        expressions.append(_tmp_e)
                        continue
                    if _tmp_e.is_expression_type(ExpressionType.LOGICAL):
                        propositions.append(_tmp_e)
                        continue
                except AttributeError:
                    pass
                # Note: Constraint.(In)Feasible is now a full relational
                # expression, so we don't need to check for it.
                # LogicalConstraint.(In)Feasible is a BooleanConstant,
                # so is not an expression and needs to be explicitly
                # checked for.
                if _tmp_e in (LogicalConstraint.Feasible, LogicalConstraint.Infeasible):
                    propositions.append(_tmp_e)
                    continue
                msg = " in '%s'" % (type(e).__name__,) if e_iter is e else ""
                raise ValueError(
                    "Unexpected term for Disjunction '%s'.\n    "
                    "Expected a Disjunct object, relational or logical "
                    "expression, or\n    iterable of relational/logical "
                    "expressions but got '%s'%s"
                    % (self.name, type(_tmp_e).__name__, msg)
                )

            comp = self.parent_component()
            if comp._autodisjuncts is None:
                b = self.parent_block()
                comp._autodisjuncts = Disjunct(Any)
                b.add_component(
                    unique_component_name(b, comp.local_name + "_disjuncts"),
                    comp._autodisjuncts,
                )
                # TODO: I am not at all sure why we need to
                # explicitly construct this block - that should
                # happen automatically.
                comp._autodisjuncts.construct()
            disjunct = comp._autodisjuncts[len(comp._autodisjuncts)]
            disjunct.constraint = ConstraintList(rule=expressions)
            disjunct.propositions = LogicalConstraintList(rule=propositions)
            self.disjuncts.append(disjunct)


class _DisjunctionData(metaclass=RenamedClass):
    __renamed__new_class__ = DisjunctionData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("Disjunction expressions.")
class Disjunction(ActiveIndexedComponent):
    _ComponentDataClass = DisjunctionData

    def __new__(cls, *args, **kwds):
        if cls != Disjunction:
            return super(Disjunction, cls).__new__(cls)
        if args == ():
            return ScalarDisjunction.__new__(ScalarDisjunction)
        else:
            return IndexedDisjunction.__new__(IndexedDisjunction)

    def __init__(self, *args, **kwargs):
        self._init_rule = kwargs.pop('rule', None)
        self._init_expr = kwargs.pop('expr', None)
        self._init_xor = _Initializer.process(kwargs.pop('xor', True))
        self._autodisjuncts = None
        kwargs.setdefault('ctype', Disjunction)
        super(Disjunction, self).__init__(*args, **kwargs)

        if self._init_expr is not None and self._init_rule is not None:
            raise ValueError(
                "Cannot specify both rule= and expr= for Disjunction %s" % (self.name,)
            )

    #
    # TODO: Ideally we would not override these methods and instead add
    # the contents of _check_skip_add to the set_value() method.
    # Unfortunately, until IndexedComponentData objects know their own
    # index, determining the index is a *very* expensive operation.  If
    # we refactor things so that the Data objects have their own index,
    # then we can remove these overloads.
    #

    def _setitem_impl(self, index, obj, value):
        if value is Disjunction.Skip:
            del self[index]
            return None
        else:
            obj.set_value(value)
            return obj

    def _setitem_when_not_present(self, index, value):
        if value is Disjunction.Skip:
            return None
        else:
            ans = super(Disjunction, self)._setitem_when_not_present(
                index=index, value=value
            )
            self._initialize_members((index,))
            return ans

    def _initialize_members(self, init_set):
        if self._init_xor[0] == _Initializer.value:  # POD data
            val = self._init_xor[1]
            for key in init_set:
                self._data[key].xor = val
        elif self._init_xor[0] == _Initializer.deferred_value:  # Param data
            val = bool(value(self._init_xor[1]))
            for key in init_set:
                self._data[key].xor = val
        elif self._init_xor[0] == _Initializer.function:  # rule
            fcn = self._init_xor[1]
            for key in init_set:
                self._data[key].xor = bool(
                    value(apply_indexed_rule(self, fcn, self._parent(), key))
                )
        elif self._init_xor[0] == _Initializer.dict_like:  # dict-like thing
            val = self._init_xor[1]
            for key in init_set:
                self._data[key].xor = bool(value(val[key]))

    def construct(self, data=None):
        if is_debug_set(logger):
            logger.debug("Constructing disjunction %s" % (self.name))
        if self._constructed:
            return
        timer = ConstructionTimer(self)
        self._constructed = True

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        _self_parent = self.parent_block()
        if not self.is_indexed():
            if self._init_rule is not None:
                expr = self._init_rule(_self_parent)
            elif self._init_expr is not None:
                expr = self._init_expr
            else:
                timer.report()
                return

            if expr is None:
                raise ValueError(_rule_returned_none_error % (self.name,))
            if expr is Disjunction.Skip:
                timer.report()
                return
            self._data[None] = self
            self._setitem_when_not_present(None, expr)
        elif self._init_expr is not None:
            raise IndexError(
                "Disjunction '%s': Cannot initialize multiple indices "
                "of a disjunction with a single disjunction list" % (self.name,)
            )
        elif self._init_rule is not None:
            _init_rule = self._init_rule
            for ndx in self._index_set:
                try:
                    expr = apply_indexed_rule(self, _init_rule, _self_parent, ndx)
                except Exception:
                    err = sys.exc_info()[1]
                    logger.error(
                        "Rule failed when generating expression for "
                        "disjunction %s with index %s:\n%s: %s"
                        % (self.name, str(ndx), type(err).__name__, err)
                    )
                    raise
                if expr is None:
                    _name = "%s[%s]" % (self.name, str(ndx))
                    raise ValueError(_rule_returned_none_error % (_name,))
                if expr is Disjunction.Skip:
                    continue
                self._setitem_when_not_present(ndx, expr)
        timer.report()

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
                ("Active", self.active),
            ],
            self.items,
            ("Disjuncts", "Active", "XOR"),
            lambda k, v: [[x.name for x in v.disjuncts], v.active, v.xor],
        )


class ScalarDisjunction(DisjunctionData, Disjunction):
    def __init__(self, *args, **kwds):
        DisjunctionData.__init__(self, component=self)
        Disjunction.__init__(self, *args, **kwds)
        self._index = UnindexedComponent_index

    #
    # Singleton disjunctions are strange in that we want them to be
    # both be constructed but have len() == 0 when not initialized with
    # anything (at least according to the unit tests that are
    # currently in place). So during initialization only, we will
    # treat them as "indexed" objects where things like
    # Constraint.Skip are managed. But after that they will behave
    # like DisjunctionData objects where set_value does not handle
    # Disjunction.Skip but expects a valid expression or None.
    #

    def set_value(self, expr):
        """Set the expression on this disjunction."""
        if not self._constructed:
            raise ValueError(
                "Setting the value of disjunction '%s' "
                "before the Disjunction has been constructed (there "
                "is currently no object to set)." % (self.name)
            )

        if len(self._data) == 0:
            self._data[None] = self
        if expr is Disjunction.Skip:
            del self[None]
            return None
        return super(ScalarDisjunction, self).set_value(expr)


class SimpleDisjunction(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarDisjunction
    __renamed__version__ = '6.0'


class IndexedDisjunction(Disjunction):
    #
    # HACK: this should be implemented on ActiveIndexedComponent, but
    # that will take time and a PEP
    #
    @property
    def active(self):
        return any(d.active for d in self._data.values())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Test data model items for QTreeView. These tests should work even without PyQt.
"""

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    Var,
    BooleanVar,
    Block,
    Param,
    Expression,
    Constraint,
    Objective,
    ExternalFunction,
    Reals,
    log,
    sin,
    sqrt,
    expr,
)
import pyomo.environ as pyo
from pyomo.contrib.viewer.model_browser import ComponentDataItem
from pyomo.contrib.viewer.ui_data import UIData
from pyomo.common.dependencies import DeferredImportError
from pyomo.core.base.units_container import pint_available


@unittest.skipIf(not pint_available, "Pyomo units are not available")
class TestDataModelItem(unittest.TestCase):
    def setUp(self):
        # Borrowed this test model from the trust region tests
        m = ConcreteModel()
        m.y = BooleanVar(range(3), initialize=False)
        m.z = Var(range(3), domain=Reals, initialize=2.0, units=pyo.units.m)
        m.x = Var(range(4), initialize=2.0, units=pyo.units.m)
        m.x[1] = 1.0
        m.x[2] = 0.0
        m.x[3] = None

        m.b1 = Block()
        m.b1.e1 = Expression(expr=m.x[0] + m.x[1])
        m.b1.e2 = Expression(expr=m.x[0] / m.x[2])
        m.b1.e3 = Expression(expr=m.x[3] * m.x[1])
        m.b1.e4 = Expression(expr=log(m.x[2]))
        m.b1.e5 = Expression(expr=log(m.x[2] - 2))

        def blackbox(a, b):
            return sin(a - b)

        self.bb = ExternalFunction(blackbox)

        m.obj = Objective(
            expr=(m.z[0] - 1.0) ** 2
            + (m.z[0] - m.z[1]) ** 2
            + (m.z[2] - 1.0) ** 2
            + (m.x[0] - 1.0) ** 4
            + (m.x[1] - 1.0) ** 6  # + m.bb(m.x[0],m.x[1])
        )
        m.c1 = Constraint(
            expr=m.x[0] * m.z[0] ** 2 + self.bb(m.x[0], m.x[1]) == 2 * sqrt(2.0)
        )
        m.c2 = Constraint(expr=m.z[2] ** 4 * m.z[1] ** 2 + m.z[1] == 8 + sqrt(2.0))
        m.c3 = Constraint(expr=m.x[1] == 3)
        m.c4 = Constraint(expr=0 == 3 / m.x[2])
        m.c5 = Constraint(expr=0 == log(m.x[2]))
        m.c6 = Constraint(expr=0 == log(m.x[2] - 4))
        m.c7 = Constraint(expr=0 == log(m.x[3]))
        m.p1 = Param(mutable=True, initialize=1)
        m.p2 = Param(initialize=1)
        m.p3 = Param(initialize=3.2)
        m.p4 = Param([1, 2, 3], mutable=True, initialize=1)
        m.c8 = Constraint(expr=m.x[1] <= 1 / m.p1)
        m.p1 = 0
        self.m = m.clone()

    def test_expr_calc(self):
        cdi = ComponentDataItem(
            parent=None, ui_data=UIData(model=self.m), o=self.m.b1.e1
        )
        cdi.ui_data.calculate_expressions()
        self.assertAlmostEqual(cdi.get("value"), 3)
        self.assertIsInstance(cdi.get("expr"), str)  # test get expr str

    def test_expr_calc_div0(self):
        cdi = ComponentDataItem(
            parent=None, ui_data=UIData(model=self.m), o=self.m.b1.e2
        )
        cdi.ui_data.calculate_expressions()
        self.assertEqual(cdi.get("value"), "Divide_by_0")

    def test_expr_calc_log0(self):
        cdi = ComponentDataItem(
            parent=None, ui_data=UIData(model=self.m), o=self.m.b1.e4
        )
        cdi.ui_data.calculate_expressions()
        self.assertIsNone(cdi.get("value"))

    def test_expr_calc_log_neg(self):
        cdi = ComponentDataItem(
            parent=None, ui_data=UIData(model=self.m), o=self.m.b1.e5
        )
        cdi.ui_data.calculate_expressions()
        self.assertIsNone(cdi.get("value"))

    def test_expr_calc_value_None(self):
        cdi = ComponentDataItem(
            parent=None, ui_data=UIData(model=self.m), o=self.m.b1.e3
        )
        cdi.ui_data.calculate_expressions()
        self.assertIsNone(cdi.get("value"))

    def test_cons_calc(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c3)
        cdi.ui_data.calculate_constraints()
        self.assertAlmostEqual(cdi.get("residual"), 2)

    def test_cons_calc_div0(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c4)
        cdi.ui_data.calculate_constraints()
        self.assertEqual(cdi.get("value"), "Divide_by_0")

    def test_cons_calc_log0(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c5)
        cdi.ui_data.calculate_constraints()
        self.assertIsNone(cdi.get("value"))

    def test_cons_calc_log_neg(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c6)
        cdi.ui_data.calculate_constraints()
        self.assertIsNone(cdi.get("value"))

    def test_cons_calc_value_None(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c7)
        cdi.ui_data.calculate_constraints()
        self.assertIsNone(cdi.get("value"))

    def test_cons_calc_upper_div0(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c8)
        cdi.ui_data.calculate_constraints()
        # the ui lists the upper and lower attributes as ub and lb
        # this was originally so I could easily combine variables and
        # constraint in the same view, but I split them up, so may want
        # to reconsider that choice in the future. This is to remind myself
        # why I'm getting "ub" and not "upper"
        self.assertEqual(cdi.get("ub"), "Divide_by_0")

    def test_cons_calc_lower(self):
        cdi = ComponentDataItem(parent=None, ui_data=UIData(model=self.m), o=self.m.c7)
        cdi.ui_data.calculate_constraints()
        # the ui lists the upper and lower attributes as ub and lb
        # this was originally so I could easily combine variables and
        # constraint in the same view, but I split them up, so may want
        # to reconsider that choice in the future. This is to remind myself
        # why I'm getting "ub" and not "upper"
        self.assertEqual(cdi.get("lb"), 0)

    def test_var_get_value(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.assertAlmostEqual(cdi.get("value"), 1)
        self.assertIsNone(cdi.get(expr))  # test can't get expr

    def test_var_get_units(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.assertAlmostEqual(cdi.get("units"), "m")
        self.assertIsNone(cdi.get(expr))  # test can't get expr

    def test_var_get_bounds(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.m.x[1].setlb(0)
        self.m.x[1].setub(10)
        self.assertAlmostEqual(cdi.get("lb"), 0)
        self.assertAlmostEqual(cdi.get("ub"), 10)

    def test_var_set_bounds(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        cdi.set("lb", 2)
        cdi.set("ub", 8)
        self.assertAlmostEqual(cdi.get("lb"), 2)
        self.assertAlmostEqual(cdi.get("ub"), 8)
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x)
        cdi.set("lb", 0)
        cdi.set("ub", 10)
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.assertAlmostEqual(cdi.get("lb"), 0)
        self.assertAlmostEqual(cdi.get("ub"), 10)

    def test_var_fixed_bounds(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        cdi.set("fixed", True)
        self.assertTrue(cdi.get("fixed"))
        cdi.set("fixed", False)
        self.assertFalse(cdi.get("fixed"))

    def test_get_attr_that_does_not_exist(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.assertIsNone(cdi.get("test_val"))

    def test_set_func(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x[1])
        self.assertIsNone(cdi.set("test_val", 5))
        self.assertIsNone(cdi.get("test_val"))  # test can't set
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.x)
        self.assertIsNone(cdi.set("test_val", 5))
        self.assertEqual(cdi.get("test_val"), 5)  # test can set with no callback

    def test_get_set_param(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.p1)
        self.assertIsNone(cdi.set("value", 5))
        self.assertEqual(cdi.get("value"), 5)
        self.assertIsNone(cdi.set("value", 0))
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.p2)
        self.assertEqual(cdi.get("value"), 1)
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.p3)
        self.assertEqual(cdi.get("value"), 3.2)
        self.assertEqual(cdi.get("expr"), None)
        self.assertIsNone(cdi.set("value", 5))
        self.assertEqual(cdi.get("value"), 3.2)
        self.assertEqual(cdi.get("units"), "dimensionless")
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.p4)
        self.assertIsNone(cdi.set("value", 6))
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.p4[1])
        self.assertEqual(cdi.get("value"), 6)

    def test_get_set_boolvar(self):
        cdi = ComponentDataItem(parent=None, ui_data=None, o=self.m.y)
        self.assertIsNone(cdi.set("value", True))
        self.assertEqual(pyo.value(self.m.y[1]), True)
        self.assertIsNone(cdi.set("value", False))
        self.assertEqual(pyo.value(self.m.y[1]), False)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly reclassifies nonlinear expressions
#          as linear or trivial when fixing variables or params
#          cause such a situation.
#
#          This test model relies on the gjh_asl_json executable. It
#          will not solve if sent to a real optimizer.
#

from pyomo.environ import (
    ConcreteModel,
    Var,
    Param,
    Objective,
    Constraint,
    simple_constraint_rule,
)

model = ConcreteModel()

model.x = Var()
model.y = Var(initialize=0.0)
model.z = Var()
model.p = Param(initialize=0.0, mutable=True)
model.q = Param(initialize=0.0, mutable=False)

model.y.fixed = True

model.obj = Objective(expr=model.x)

model.con1 = Constraint(expr=model.x * model.y * model.z + model.x == 1.0)
model.con2 = Constraint(expr=model.x * model.p * model.z + model.x == 1.0)
model.con3 = Constraint(expr=model.x * model.q * model.z + model.x == 1.0)
# Pyomo differs from AMPL in these cases that involve immutable params (q).
# These never actually become constants in Pyomo, and for good reason.
model.con4 = Constraint(expr=model.x * model.y * model.z == 1.0)
model.con5 = Constraint(expr=model.x * model.p * model.z == 1.0)
model.con6 = Constraint(rule=simple_constraint_rule(model.x * model.q * model.z == 0.0))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import platform
import math
import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.common.fileutils import find_library

flib = find_library("aslfunctions")
is_pypy = platform.python_implementation().lower().startswith("pypy")


@unittest.skipIf(not flib, 'Could not find the "aslfunctions" library')
@unittest.skipIf(is_pypy, "Cannot evaluate external functions under pypy")
class TestAMPLExternalFunction(unittest.TestCase):

    def test_eval_sqnsqr_function_fgh(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="sgnsqr")

        f, g, h = m.tf.evaluate_fgh((2,))
        self.assertEqual(f, 4)
        self.assertEqual(g, [4])
        self.assertEqual(h, [2])

        f, g, h = m.tf.evaluate_fgh((-2,))
        self.assertAlmostEqual(f, -4)
        self.assertStructuredAlmostEqual(g, [4])
        self.assertStructuredAlmostEqual(h, [-2])

    def test_eval_sqnsqr_c4_function_fgh(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="sgnsqr_c4")

        f, g, h = m.tf.evaluate_fgh((2,))
        self.assertEqual(f, 4)
        self.assertEqual(g, [4])
        self.assertEqual(h, [2])

        f, g, h = m.tf.evaluate_fgh((-2,))
        self.assertAlmostEqual(f, -4)
        self.assertStructuredAlmostEqual(g, [4])
        self.assertStructuredAlmostEqual(h, [-2])

        dx = 1e-9
        x = 0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)

        dx = 1e-9
        x = -0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)

    def test_eval_sinc_function_fgh(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="sinc")

        f, g, h = m.tf.evaluate_fgh((0,))
        self.assertAlmostEqual(f, 1.0)
        self.assertAlmostEqual(g, [0])
        self.assertAlmostEqual(h, [-1 / 3.0])

        dx = 1e-10
        f, g, h = m.tf.evaluate_fgh((2,))
        self.assertAlmostEqual(f, math.sin(2) / 2)
        self.assertStructuredAlmostEqual(g, [(math.sin(2 + dx) / (2 + dx) - f) / dx], 5)

        dx = 1e-10
        x = 0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)

        dx = 1e-10
        x = 0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)

    def test_eval_sqnsqrt_c4_function_fgh(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="sgnsqrt_c4")

        f, g, h = m.tf.evaluate_fgh((4,))
        self.assertEqual(f, 2)
        self.assertEqual(g, [0.5 * 4 ** (-0.5)])
        self.assertEqual(h, [-0.25 * 4 ** (-3 / 2)])

        f, g, h = m.tf.evaluate_fgh((-4,))
        self.assertAlmostEqual(f, -2)
        self.assertStructuredAlmostEqual(g, [0.5 * 4 ** (-0.5)])
        self.assertStructuredAlmostEqual(h, [0.25 * 4 ** (-3 / 2)])

        dx = 1e-9
        x = 0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)

        dx = 1e-9
        x = -0.1

        f1, g1, h1 = m.tf.evaluate_fgh((x + dx,))
        f2, g2, h2 = m.tf.evaluate_fgh((x - dx,))
        self.assertAlmostEqual(f1, f2)
        self.assertStructuredAlmostEqual(g1, g2)
        self.assertStructuredAlmostEqual(h1, h2)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.opt import (
    TerminationCondition,
    SolutionStatus,
    SolverStatus,
    check_available_solvers,
)
import pyomo.environ as pyo
import pyomo.kernel as pmo
import sys

diff_tol = 1e-3

mosek_available = check_available_solvers('mosek_direct')
msk_version = [0]
if mosek_available:
    msk_version = pyo.SolverFactory('mosek')._version


@unittest.skipIf(not mosek_available, "MOSEK's python bindings are missing.")
@unittest.pytest.mark.solver("mosek_persistent")
class MOSEKPersistentTests(unittest.TestCase):
    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None

    def tearDown(self):
        sys.stderr = self.stderr

    def test_interface_call(self):
        interface_instance = type(pyo.SolverFactory('mosek_persistent'))
        alt_1 = pyo.SolverFactory('mosek', solver_io='persistent')
        self.assertIsInstance(alt_1, interface_instance)

    def test_variable_removal(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        opt = pyo.SolverFactory('mosek_persistent')
        opt.set_instance(m)

        self.assertEqual(opt._solver_model.getnumvar(), 2)

        opt.remove_var(m.x)
        self.assertEqual(opt._solver_model.getnumvar(), 1)

        opt.remove_var(m.y)
        self.assertEqual(opt._solver_model.getnumvar(), 0)
        self.assertRaises(ValueError, opt.remove_var, m.x)

        opt.add_var(m.x)
        opt.add_var(m.y)
        self.assertEqual(opt._solver_model.getnumvar(), 2)
        opt.remove_vars(m.x, m.y)
        self.assertEqual(opt._solver_model.getnumvar(), 0)

    def test_constraint_removal_1(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.c1 = pyo.Constraint(expr=2 * m.x >= m.y**2)
        m.c2 = pyo.Constraint(expr=m.x**2 >= m.y**2 + m.z**2)
        m.c3 = pyo.Constraint(expr=m.z >= 0)
        m.c4 = pyo.Constraint(expr=m.x + m.y >= 0)
        opt = pyo.SolverFactory('mosek_persistent')
        opt.set_instance(m)

        self.assertEqual(opt._solver_model.getnumcon(), 4)

        opt.remove_constraint(m.c1)
        self.assertEqual(opt._solver_model.getnumcon(), 3)

        opt.remove_constraints(m.c2, m.c3)
        self.assertEqual(opt._solver_model.getnumcon(), 1)

        opt.add_constraint(m.c1)
        self.assertEqual(opt._solver_model.getnumcon(), 2)
        self.assertRaises(ValueError, opt.remove_constraint, m.c2)

    @unittest.skipIf(
        msk_version[0] > 9, "MOSEK 10 does not (yet) have a removeacc method."
    )
    def test_constraint_removal_2(self):
        m = pmo.block()
        m.x = pmo.variable()
        m.y = pmo.variable()
        m.z = pmo.variable()
        m.c1 = pmo.conic.rotated_quadratic.as_domain(2, m.x, [m.y])
        m.c2 = pmo.conic.quadratic(m.x, [m.y, m.z])
        m.c3 = pmo.constraint(m.z >= 0)
        m.c4 = pmo.constraint(m.x + m.y >= 0)
        opt = pmo.SolverFactory('mosek_persistent')
        opt.set_instance(m)

        self.assertEqual(opt._solver_model.getnumcon(), 5)
        self.assertEqual(opt._solver_model.getnumcone(), 2)

        opt.remove_block(m.c1)
        self.assertEqual(opt._solver_model.getnumcon(), 2)
        self.assertEqual(opt._solver_model.getnumcone(), 1)

        opt.remove_constraints(m.c2, m.c3)
        self.assertEqual(opt._solver_model.getnumcon(), 1)
        self.assertEqual(opt._solver_model.getnumcone(), 0)

        self.assertRaises(ValueError, opt.remove_constraint, m.c2)
        opt.add_constraint(m.c2)
        opt.add_block(m.c1)
        self.assertEqual(opt._solver_model.getnumcone(), 2)

    def test_column_addition(self):
        '''
        Test based on lo1.py problem from MOSEK documentation.
        '''
        sol_to_get = [0.0, 0.0, 15.0, 8]

        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, None))
        m.y = pyo.Var(bounds=(0, 10))
        m.z = pyo.Var(bounds=(0, None))
        m.c1 = pyo.Constraint(expr=3 * m.x + m.y + 2 * m.z == 30)
        m.c2 = pyo.Constraint(expr=2 * m.x + m.y + 3 * m.z >= 15)
        m.c3 = pyo.Constraint(expr=2 * m.y <= 25)
        m.o = pyo.Objective(expr=3 * m.x + m.y + 5 * m.z, sense=pyo.maximize)
        opt = pyo.SolverFactory('mosek_persistent')
        opt.set_instance(m)

        m.new_var = pyo.Var(bounds=(0, None))
        opt.add_column(m, m.new_var, 1, [m.c2, m.c3], [1, 3])

        self.assertEqual(opt._solver_model.getnumvar(), 4)

        opt.solve(m)
        for i, v in enumerate([m.x, m.y, m.z, m.new_var]):
            with self.subTest(i=v.name):
                self.assertAlmostEqual(v.value, sol_to_get[i], places=0)

    def test_variable_update(self):
        '''
        Test based on milo1.py problem from MOSEK documentation.
        '''
        cont_sol_to_get = [1.948, 4.922]
        int_sol_to_get = [5, 0]

        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c1 = pyo.Constraint(expr=50 * m.x + 31 * m.y <= 250)
        m.c2 = pyo.Constraint(expr=3 * m.x - 2 * m.y >= -4)
        m.o = pyo.Objective(expr=m.x + 0.64 * m.y, sense=pyo.maximize)
        opt = pyo.SolverFactory('mosek_persistent')
        opt.set_instance(m)
        opt.solve(m)

        self.assertAlmostEqual(m.x.value, cont_sol_to_get[0], places=2)
        self.assertAlmostEqual(m.y.value, cont_sol_to_get[1], places=2)

        m.x.setlb = 0
        m.x.setub = None
        m.x.domain = pyo.Integers
        m.y.setlb = 0
        m.y.setub = None
        m.y.domain = pyo.Integers
        m.z = pyo.Var()

        opt.update_vars(m.x, m.y)
        self.assertRaises(ValueError, opt.update_var, m.z)
        opt.add_var(m.z)

        opt.solve(m)
        self.assertAlmostEqual(m.x.value, int_sol_to_get[0], places=1)
        self.assertAlmostEqual(m.y.value, int_sol_to_get[1], places=1)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo.environ as pyo

from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
    RooneyBieglerExperiment,
)


class BadExperiment:
    def __init__(self):
        self.model = None


class RooneyBieglerExperimentFlag(RooneyBieglerExperiment):
    """
    A version of RooneyBieglerExperiment that supports flag parameter
    to create incomplete models for testing error handling.

    Flag values:
    - 0: Full model (all suffixes present)
    - 1: No experiment_outputs suffix
    - 2: No measurement_error suffix
    - 3: No experiment_inputs suffix
    - 4: No unknown_parameters suffix
    - 5: Incorrect measurement error structure (only one output gets error)
    """

    def get_labeled_model(self, flag=0):
        """Get the labeled model with optional flag to create incomplete models."""
        m = super().get_labeled_model()

        # If flag is non-zero, remove the corresponding suffix to create an incomplete model
        if flag == 1:
            m.del_component(m.experiment_outputs)
        elif flag == 2:
            m.del_component(m.measurement_error)
        elif flag == 3:
            m.del_component(m.experiment_inputs)
        elif flag == 4:
            m.del_component(m.unknown_parameters)
        elif flag == 5:
            # Create mismatch: 1 experiment output but multiple measurement errors
            # This tests the validation that checks output/error length match
            # Add a fake component to measurement_error to create length mismatch
            m.fake_output = pyo.Var()
            m.measurement_error[m.fake_output] = 0.1

        return m


class RooneyBieglerExperimentBad(RooneyBieglerExperiment):
    """
    A bad version of RooneyBieglerExperiment with conflicting constraints
    for testing error handling.
    """

    def get_labeled_model(self):
        m = super().get_labeled_model()

        # Add conflicting constraints that make the model infeasible
        # The hour variable should be >= 10 and <= 0 (impossible)
        m.bad_con_1 = pyo.Constraint(expr=m.hour >= 10.0)
        m.bad_con_2 = pyo.Constraint(expr=m.hour <= 0.0)

        return m



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for components
#
from io import StringIO
import pyomo.common.unittest as unittest

from pyomo.common import DeveloperError
from pyomo.environ import (
    ConcreteModel,
    Component,
    Block,
    Var,
    Set,
    ModelComponentFactory,
)
from pyomo.core.base.set import GlobalSets


class TestComponent(unittest.TestCase):
    def test_construct_component_throws_exception(self):
        with self.assertRaisesRegex(
            DeveloperError, "Must specify a component type for class Component"
        ):
            Component()

    def test_getname(self):
        m = ConcreteModel()
        m.b = Block([1, 2])
        m.b[2].c = Var([1, 2], [3, 4])

        self.assertEqual(m.getname(fully_qualified=True), "unknown")
        self.assertEqual(m.b.getname(fully_qualified=True), "b")
        self.assertEqual(m.b[1].getname(fully_qualified=True), "b[1]")
        self.assertEqual(m.b[2].c[2, 4].getname(fully_qualified=True), "b[2].c[2,4]")
        self.assertEqual(m.getname(fully_qualified=False), "unknown")
        self.assertEqual(m.b.getname(fully_qualified=False), "b")
        self.assertEqual(m.b[1].getname(fully_qualified=False), "b[1]")
        self.assertEqual(m.b[2].c[2, 4].getname(fully_qualified=False), "c[2,4]")

        cache = {}
        self.assertEqual(
            m.b[2].c[2, 4].getname(fully_qualified=True, name_buffer=cache),
            "b[2].c[2,4]",
        )
        self.assertEqual(len(cache), 8)
        self.assertIn(id(m.b[2].c[2, 4]), cache)
        self.assertIn(id(m.b[2].c[1, 3]), cache)
        self.assertIn(id(m.b[2].c), cache)
        self.assertIn(id(m.b[2]), cache)
        self.assertIn(id(m.b[1]), cache)
        self.assertIn(id(m.b), cache)
        self.assertNotIn(id(m), cache)
        self.assertEqual(
            m.b[2].c[1, 3].getname(fully_qualified=True, name_buffer=cache),
            "b[2].c[1,3]",
        )

        m.b[2]._component = None
        self.assertEqual(m.b[2].getname(fully_qualified=True), "[Unattached BlockData]")
        # I think that getname() should do this:
        # self.assertEqual(m.b[2].c[2,4].getname(fully_qualified=True),
        #                 "[Unattached BlockData].c[2,4]")
        # but it doesn't match current behavior.  I will file a PEP to
        # propose changing the behavior later and proceed to test
        # current behavior.
        self.assertEqual(m.b[2].c[2, 4].getname(fully_qualified=True), "c[2,4]")

        self.assertEqual(
            m.b[2].getname(fully_qualified=False), "[Unattached BlockData]"
        )
        self.assertEqual(m.b[2].c[2, 4].getname(fully_qualified=False), "c[2,4]")

        # Cached names still work...
        self.assertEqual(
            m.b[2].getname(fully_qualified=True, name_buffer=cache), "b[2]"
        )
        self.assertEqual(
            m.b[2].c[1, 3].getname(fully_qualified=True, name_buffer=cache),
            "b[2].c[1,3]",
        )

    def test_component_data_pprint(self):
        m = ConcreteModel()
        m.a = Set(initialize=[1, 2, 3], ordered=True)
        m.x = Var(m.a)
        stream = StringIO()
        m.x[2].pprint(ostream=stream)
        correct_s = (
            '{Member of x} : Size=3, Index=a\n    '
            'Key : Lower : Value : Upper : Fixed : Stale : Domain\n      '
            '2 :  None :  None :  None : False :  True :  Reals\n'
        )
        self.assertEqual(correct_s, stream.getvalue())

    def test_is_reference(self):
        m = ConcreteModel()

        class _NotSpecified:
            pass

        m.comp = Component(ctype=_NotSpecified)
        self.assertFalse(m.comp.is_reference())


class TestEnviron(unittest.TestCase):
    def test_components(self):
        self.assertGreaterEqual(
            set(ModelComponentFactory),
            set(['Set', 'Param', 'Var', 'Objective', 'Constraint']),
        )

    def test_sets(self):
        self.assertGreaterEqual(set(GlobalSets), set(['Reals', 'Integers', 'Boolean']))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For testing to ensure that the Pyomo NL writer properly
#          handles the Expr_if component.
#
#          This test model relies on the gjh_asl_json executable. It
#          will not solve if sent to a real optimizer.
#

from pyomo.environ import ConcreteModel, Var, Param, Objective, Constraint, inequality
from pyomo.core.expr import Expr_if

model = ConcreteModel()

model.vTrue = Var(initialize=1)
model.vFalse = Var(initialize=-1)

model.pTrue = Param(initialize=1)
model.pFalse = Param(initialize=-1)

model.vN1 = Var(initialize=-1)
model.vP1 = Var(initialize=1)
model.v0 = Var(initialize=0)
model.vN2 = Var(initialize=-2)
model.vP2 = Var(initialize=2)

model.obj = Objective(
    expr=10.0 * Expr_if(IF=model.v0, THEN=model.vTrue, ELSE=model.vFalse)
)

# True/False
model.c1 = Constraint(
    expr=Expr_if(IF=(0), THEN=(model.vTrue), ELSE=(model.vFalse)) == model.pFalse
)
model.c2 = Constraint(
    expr=Expr_if(IF=(1), THEN=(model.vTrue), ELSE=(model.vFalse)) == model.pTrue
)

# x <= 0
model.c3 = Constraint(
    expr=Expr_if(IF=(model.vN1 <= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)
model.c4 = Constraint(
    expr=Expr_if(IF=(model.v0 <= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)
model.c5 = Constraint(
    expr=Expr_if(IF=(model.vP1 <= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)

# x < 0
model.c6 = Constraint(
    expr=Expr_if(IF=(model.vN1 < 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)
model.c7 = Constraint(
    expr=Expr_if(IF=(model.v0 < 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)
model.c8 = Constraint(
    expr=Expr_if(IF=(model.vP1 < 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)

# x >= 0
model.c9 = Constraint(
    expr=Expr_if(IF=(model.vN1 * 10.0 >= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)
model.c10 = Constraint(
    expr=Expr_if(IF=(model.v0 * 10.0 >= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)
model.c11 = Constraint(
    expr=Expr_if(IF=(model.vP1 * 10.0 >= 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)

# x > 0
model.c12 = Constraint(
    expr=Expr_if(IF=(model.vN1 * 10.0 > 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)
model.c13 = Constraint(
    expr=Expr_if(IF=(model.v0 * 10.0 > 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pFalse
)
model.c14 = Constraint(
    expr=Expr_if(IF=(model.vP1 * 10.0 > 0), THEN=(model.vTrue), ELSE=(model.vFalse))
    == model.pTrue
)

# -1 <= x <= 1
model.c15 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vN2, 1), THEN=(model.vTrue), ELSE=(model.vFalse)
    )
    == model.pFalse
)
model.c16 = Constraint(
    expr=Expr_if(
        IF=inequality(-1 * model.vP1, model.vN1, 1),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pTrue
)
model.c17 = Constraint(
    expr=Expr_if(
        IF=inequality(-1 * model.vP1**2, model.v0, 1),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pTrue
)
model.c18 = Constraint(
    expr=Expr_if(
        IF=inequality(model.vN1, model.vP1, 1), THEN=(model.vTrue), ELSE=(model.vFalse)
    )
    == model.pTrue
)
model.c19 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vP2, 1), THEN=(model.vTrue), ELSE=(model.vFalse)
    )
    == model.pFalse
)

# -1 < x < 1
model.c20 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vN2, 1, strict=True),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pFalse
)
model.c21 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vN1, 1 * model.vP1, strict=True),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pFalse
)
model.c22 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.v0, 1 * model.vP1**2, strict=True),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pTrue
)
model.c23 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vP1, model.vP1, strict=True),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pFalse
)
model.c24 = Constraint(
    expr=Expr_if(
        IF=inequality(-1, model.vP2, 1, strict=True),
        THEN=(model.vTrue),
        ELSE=(model.vFalse),
    )
    == model.pFalse
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Utility functions and a utility class for interfacing Pyomo components with
useful graph algorithms.

"""

import enum
import textwrap
from pyomo.core.base.block import BlockData
from pyomo.core.base.var import Var
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.objective import Objective
from pyomo.core.expr import EqualityExpression
from pyomo.util.subsystems import create_subsystem_block
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.dependencies import (
    attempt_import,
    networkx as nx,
    scipy as sp,
    plotly,
)
from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.contrib.incidence_analysis.config import get_config_from_kwds
from pyomo.contrib.incidence_analysis.matching import maximum_matching
from pyomo.contrib.incidence_analysis.connected import get_independent_submatrices
from pyomo.contrib.incidence_analysis.triangularize import (
    get_scc_of_projection,
    block_triangularize,
    get_diagonal_blocks,
    get_blocks_from_maps,
)
from pyomo.contrib.incidence_analysis.dulmage_mendelsohn import (
    dulmage_mendelsohn,
    RowPartition,
    ColPartition,
)
from pyomo.contrib.incidence_analysis.incidence import get_incident_variables
from pyomo.contrib.pynumero.asl import AmplInterface

pyomo_nlp, pyomo_nlp_available = attempt_import(
    "pyomo.contrib.pynumero.interfaces.pyomo_nlp"
)
asl_available = pyomo_nlp_available & AmplInterface.available()


def _check_unindexed(complist):
    for comp in complist:
        if comp.is_indexed():
            raise RuntimeError(
                "Variables and constraints must be unindexed "
                "ComponentData objects. Got %s, which is indexed." % comp.name
            )


def get_incidence_graph(variables, constraints, **kwds):
    config = get_config_from_kwds(**kwds)
    return get_bipartite_incidence_graph(variables, constraints, **config)


def get_bipartite_incidence_graph(variables, constraints, **kwds):
    """Return the bipartite incidence graph of Pyomo variables and constraints.

    Each node in the returned graph is an integer. The convention is that,
    for a graph with N variables and M constraints, nodes 0 through M-1
    correspond to constraints and nodes M through M+N-1 correspond to variables.
    Nodes correspond to variables and constraints in the provided orders.
    For consistency with NetworkX's "convention", constraint nodes are tagged
    with ``bipartite=0`` while variable nodes are tagged with ``bipartite=1``,
    although these attributes are not used.

    Parameters
    ---------
    variables: List of Pyomo VarData objects
        Variables that will appear in incidence graph
    constraints: List of Pyomo ConstraintData objects
        Constraints that will appear in incidence graph
    include_fixed: Bool
        Flag for whether fixed variable should be included in the incidence

    Returns
    -------
    ``networkx.Graph``

    """
    # Note that this ConfigDict contains the visitor that we will re-use
    # when constructing constraints.
    config = get_config_from_kwds(**kwds)
    _check_unindexed(variables + constraints)
    N = len(variables)
    M = len(constraints)
    graph = nx.Graph()
    graph.add_nodes_from(range(M), bipartite=0)
    graph.add_nodes_from(range(M, M + N), bipartite=1)
    var_node_map = ComponentMap((v, M + i) for i, v in enumerate(variables))
    for i, con in enumerate(constraints):
        for var in get_incident_variables(con.body, **config):
            if var in var_node_map:
                graph.add_edge(i, var_node_map[var])
    return graph


def extract_bipartite_subgraph(graph, nodes0, nodes1):
    """Return the bipartite subgraph of a graph.

    Two lists of nodes to project onto must be provided. These will correspond
    to the "bipartite sets" in the subgraph. If the two sets provided have
    M and N nodes, the subgraph will have nodes 0 through M+N-1, with the first
    M corresponding to the first set provided and the last N corresponding
    to the second set.

    Parameters
    ----------
    graph: NetworkX Graph
        The graph from which a subgraph is extracted
    nodes0: list
        A list of nodes in the original graph that will form the first
        bipartite set of the projected graph (and have ``bipartite=0``)
    nodes1: list
        A list of nodes in the original graph that will form the second
        bipartite set of the projected graph (and have ``bipartite=1``)

    Returns
    -------
    subgraph: ``networkx.Graph``
        Graph containing integer nodes corresponding to positions in the
        provided lists, with edges where corresponding nodes are adjacent
        in the original graph.

    """
    subgraph = graph.subgraph(nodes0 + nodes1)
    # TODO: Any error checking that nodes are valid bipartition?
    for node in nodes0:
        bipartite = graph.nodes[node]["bipartite"]
        if bipartite != 0:
            raise RuntimeError(
                "Invalid bipartite sets. Node {node} in set 0 has"
                " bipartite={bipartite}"
            )
    for node in nodes1:
        bipartite = graph.nodes[node]["bipartite"]
        if bipartite != 1:
            raise RuntimeError(
                "Invalid bipartite sets. Node {node} in set 1 has"
                " bipartite={bipartite}"
            )
    old_new_map = {}
    for i, node in enumerate(nodes0 + nodes1):
        if node in old_new_map:
            raise RuntimeError("Node %s provided more than once.")
        old_new_map[node] = i
    relabeled_subgraph = nx.relabel_nodes(subgraph, old_new_map)
    return relabeled_subgraph


def _generate_variables_in_constraints(constraints, **kwds):
    # Note: We construct a visitor here
    config = get_config_from_kwds(**kwds)
    known_vars = ComponentSet()
    for con in constraints:
        for var in get_incident_variables(con.body, **config):
            if var not in known_vars:
                known_vars.add(var)
                yield var


def get_structural_incidence_matrix(variables, constraints, **kwds):
    """Return the incidence matrix of Pyomo constraints and variables

    Parameters
    ---------
    variables: List of Pyomo VarData objects
    constraints: List of Pyomo ConstraintData objects
    include_fixed: Bool
        Flag for whether fixed variables should be included in the matrix
        nonzeros

    Returns
    -------
    ``scipy.sparse.coo_matrix``
        COO matrix. Rows are indices into the user-provided list of constraints,
        columns are indices into the user-provided list of variables.
        Entries are 1.0.

    """
    config = get_config_from_kwds(**kwds)
    _check_unindexed(variables + constraints)
    N, M = len(variables), len(constraints)
    var_idx_map = ComponentMap((v, i) for i, v in enumerate(variables))
    rows = []
    cols = []
    for i, con in enumerate(constraints):
        cols.extend(
            var_idx_map[v]
            for v in get_incident_variables(con.body, **config)
            if v in var_idx_map
        )
        rows.extend([i] * (len(cols) - len(rows)))
    assert len(rows) == len(cols)
    data = [1.0] * len(rows)
    matrix = sp.sparse.coo_matrix((data, (rows, cols)), shape=(M, N))
    return matrix


def get_numeric_incidence_matrix(variables, constraints):
    """Return the "numeric incidence matrix" (Jacobian) of Pyomo variables
    and constraints.

    Each matrix value is the derivative of a constraint body with respect
    to a variable. Rows correspond to constraints and columns correspond to
    variables. Entries are included even if the value of the derivative is
    zero.
    Only active constraints and unfixed variables that participate in these
    constraints are included.

    Parameters
    ---------
    variables: List of Pyomo VarData objects
    constraints: List of Pyomo ConstraintData objects

    Returns
    -------
    ``scipy.sparse.coo_matrix``
        COO matrix. Rows are indices into the user-provided list of constraints,
        columns are indices into the user-provided list of variables.

    """
    # NOTE: There are several ways to get a numeric incidence matrix
    # from a Pyomo model. Here we get the numeric incidence matrix by
    # creating a temporary block and using the PyNumero ASL interface.
    comps = list(variables) + list(constraints)
    _check_unindexed(comps)
    block = create_subsystem_block(constraints, variables)
    block._obj = Objective(expr=0)
    nlp = pyomo_nlp.PyomoNLP(block)
    return nlp.extract_submatrix_jacobian(variables, constraints)


class IncidenceGraphInterface:
    """An interface for applying graph algorithms to Pyomo variables and
    constraints

    Parameters
    ----------
    model: Pyomo BlockData or PyNumero PyomoNLP, default ``None``
        An object from which an incidence graph will be constructed.
    active: Bool, default ``True``
        Whether only active constraints should be included in the incidence
        graph. Cannot be set to ``False`` if the ``model`` is provided as
        a PyomoNLP.
    include_fixed: Bool, default ``False``
        Whether to include fixed variables in the incidence graph. Cannot
        be set to ``False`` if ``model`` is a PyomoNLP.
    include_inequality: Bool, default ``True``
        Whether to include inequality constraints (those whose expressions
        are not instances of ``EqualityExpression``) in the incidence graph.
        If a PyomoNLP is provided, setting to ``False`` uses the
        ``evaluate_jacobian_eq`` method instead of ``evaluate_jacobian``
        rather than checking constraint expression types.

    """

    def __init__(self, model=None, active=True, include_inequality=True, **kwds):
        """Construct an IncidenceGraphInterface object"""
        # If the user gives us a model or an NLP, we assume they want us
        # to cache the incidence graph for fast analysis later on.
        # WARNING: This cache will become invalid if the user alters their
        # model.
        self._config = get_config_from_kwds(**kwds)
        if model is None:
            self._incidence_graph = None
            self._variables = None
            self._constraints = None
        elif isinstance(model, BlockData):
            self._constraints = [
                con
                for con in model.component_data_objects(Constraint, active=active)
                if include_inequality or isinstance(con.expr, EqualityExpression)
            ]
            self._variables = list(
                _generate_variables_in_constraints(self._constraints, **self._config)
            )
            self._var_index_map = ComponentMap(
                (var, i) for i, var in enumerate(self._variables)
            )
            self._con_index_map = ComponentMap(
                (con, i) for i, con in enumerate(self._constraints)
            )
            self._incidence_graph = get_bipartite_incidence_graph(
                self._variables, self._constraints, **self._config
            )
        elif pyomo_nlp_available and isinstance(model, pyomo_nlp.PyomoNLP):
            if not active:
                raise ValueError(
                    "Cannot get the Jacobian of inactive constraints from the "
                    "nl interface (PyomoNLP).\nPlease set the `active` flag "
                    "to True."
                )
            if kwds:
                raise ValueError(
                    "Incidence graph generation options, e.g. include_fixed, method,"
                    " and linear_only, are not supported when generating a graph"
                    " from a PyomoNLP."
                )
            nlp = model
            self._variables = nlp.get_pyomo_variables()
            self._constraints = [
                con
                for con in nlp.get_pyomo_constraints()
                if include_inequality or isinstance(con.expr, EqualityExpression)
            ]
            self._var_index_map = ComponentMap(
                (var, idx) for idx, var in enumerate(self._variables)
            )
            self._con_index_map = ComponentMap(
                (con, idx) for idx, con in enumerate(self._constraints)
            )
            if include_inequality:
                incidence_matrix = nlp.evaluate_jacobian()
            else:
                incidence_matrix = nlp.evaluate_jacobian_eq()
            nxb = nx.algorithms.bipartite
            self._incidence_graph = nxb.from_biadjacency_matrix(incidence_matrix)
        elif isinstance(model, tuple):
            # model is a tuple of (nx.Graph, list[pyo.Var], list[pyo.Constraint])
            # We could potentially accept a tuple (variables, constraints).
            # TODO: Disallow kwargs if this type of "model" is provided?
            nx_graph, variables, constraints = model
            self._variables = list(variables)
            self._constraints = list(constraints)
            self._var_index_map = ComponentMap(
                (var, i) for i, var in enumerate(self._variables)
            )
            self._con_index_map = ComponentMap(
                (con, i) for i, con in enumerate(self._constraints)
            )
            # For now, don't check any properties of this graph. We could check
            # for a bipartition that matches the variable and constraint lists.
            self._incidence_graph = nx_graph
        else:
            raise TypeError(
                "Unsupported type for incidence graph. Expected PyomoNLP"
                " or BlockData but got %s." % type(model)
            )

    @property
    def variables(self):
        """The variables participating in the incidence graph"""
        if self._incidence_graph is None:
            raise RuntimeError("Cannot get variables when nothing is cached")
        return self._variables

    @property
    def constraints(self):
        """The constraints participating in the incidence graph"""
        if self._incidence_graph is None:
            raise RuntimeError("Cannot get constraints when nothing is cached")
        return self._constraints

    @property
    def n_edges(self):
        """The number of edges in the incidence graph, or the number of
        structural nonzeros in the incidence matrix
        """
        # The number of structural nonzeros in the incidence matrix
        if self._incidence_graph is None:
            raise RuntimeError(
                "Cannot get number of edges (nonzeros) when nothing is cached"
            )
        return len(self._incidence_graph.edges)

    @property
    @deprecated(
        msg="``var_index_map`` is deprecated. Please use ``get_matrix_coord`` instead.",
        version="6.5.0",
    )
    def var_index_map(self):
        return self._var_index_map

    @property
    @deprecated(
        msg="``con_index_map`` is deprecated. Please use ``get_matrix_coord`` instead.",
        version="6.5.0",
    )
    def con_index_map(self):
        return self._con_index_map

    @property
    @deprecated(
        msg="The ``row_block_map`` attribute is deprecated and will be removed.",
        version="6.5.0",
    )
    def row_block_map(self):
        return None

    @property
    @deprecated(
        msg="The ``col_block_map`` attribute is deprecated and will be removed.",
        version="6.5.0",
    )
    def col_block_map(self):
        return None

    def get_matrix_coord(self, component):
        """Return the row or column coordinate of the component in the incidence
        *matrix* of variables and constraints

        Variables will return a column coordinate and constraints will return
        a row coordinate.

        Parameters
        ----------
        component: ``ComponentData``
            Component whose coordinate to locate

        Returns
        -------
        ``int``
            Column or row coordinate of the provided variable or constraint

        """
        if self._incidence_graph is None:
            raise RuntimeError(
                "Cannot get the coordinate of %s if an incidence graph"
                " is not cached." % component.name
            )
        _check_unindexed([component])
        if component in self._var_index_map and component in self._con_index_map:
            raise RuntimeError(
                "%s is in both variable and constraint maps."
                " This should not happen." % component.name
            )
        elif component in self._var_index_map:
            return self._var_index_map[component]
        elif component in self._con_index_map:
            return self._con_index_map[component]
        else:
            raise RuntimeError(
                "%s is not included in the incidence graph" % component.name
            )

    def _validate_input(self, variables, constraints):
        if variables is None:
            if self._incidence_graph is None:
                raise ValueError("Neither variables nor a model have been provided.")
            else:
                variables = self.variables
        elif self._incidence_graph is not None:
            # If variables were provided and an incidence graph is cached,
            # make sure the provided variables exist in the graph.
            for var in variables:
                if var not in self._var_index_map:
                    raise KeyError(
                        f"Variable {var} does not exist in the cached"
                        " incidence graph."
                    )
        if constraints is None:
            if self._incidence_graph is None:
                raise ValueError("Neither constraints nor a model have been provided.")
            else:
                constraints = self.constraints
        elif self._incidence_graph is not None:
            # If constraints were provided and an incidence graph is cached,
            # make sure the provided constraints exist in the graph.
            for con in constraints:
                if con not in self._con_index_map:
                    raise KeyError(
                        f"Constraint {con} does not exist in the cached"
                        " incidence graph."
                    )

        _check_unindexed(variables + constraints)
        return variables, constraints

    def _extract_subgraph(self, variables, constraints):
        if self._incidence_graph is None:
            # Note that we pass along self._config here, so any kwds used
            # in construction will apply to these incidence graphs.
            return get_bipartite_incidence_graph(variables, constraints, **self._config)
        else:
            constraint_nodes = [self._con_index_map[con] for con in constraints]

            # Note that this is the number of constraints in the original graph,
            # not the subgraph.
            M = len(self.constraints)
            variable_nodes = [M + self._var_index_map[var] for var in variables]
            subgraph = extract_bipartite_subgraph(
                self._incidence_graph, constraint_nodes, variable_nodes
            )
            return subgraph

    def subgraph(self, variables, constraints):
        """Extract a subgraph defined by the provided variables and constraints

        Underlying data structures are copied, and constraints are not reinspected
        for incidence variables (the edges from this incidence graph are used).

        Returns
        -------
        ``IncidenceGraphInterface``
            A new incidence graph containing only the specified variables and
            constraints, and the edges between pairs thereof.

        """
        nx_subgraph = self._extract_subgraph(variables, constraints)
        subgraph = IncidenceGraphInterface(
            (nx_subgraph, variables, constraints), **self._config
        )
        return subgraph

    @property
    def incidence_matrix(self):
        """The structural incidence matrix of variables and constraints.

        Variables correspond to columns and constraints correspond to rows.
        All matrix entries have value 1.0.

        """
        if self._incidence_graph is None:
            return None
        else:
            M = len(self.constraints)
            N = len(self.variables)
            row = []
            col = []
            data = []
            # Here we assume that the incidence graph is bipartite with nodes
            # 0 through M-1 forming one of the bipartite sets.
            for i in range(M):
                assert self._incidence_graph.nodes[i]["bipartite"] == 0
                for j in self._incidence_graph[i]:
                    assert self._incidence_graph.nodes[j]["bipartite"] == 1
                    row.append(i)
                    col.append(j - M)
                    data.append(1.0)
            return sp.sparse.coo_matrix((data, (row, col)), shape=(M, N))

    def get_adjacent_to(self, component):
        """Return a list of components adjacent to the provided component
        in the cached bipartite incidence graph of variables and constraints

        Parameters
        ----------
        component: ``ComponentData``
            The variable or constraint data object whose adjacent components
            are returned

        Returns
        -------
        list of ComponentData
            List of constraint or variable data objects adjacent to the
            provided component

        Example
        -------

        .. doctest::
           :skipif: not networkx_available

           >>> import pyomo.environ as pyo
           >>> from pyomo.contrib.incidence_analysis import IncidenceGraphInterface
           >>> m = pyo.ConcreteModel()
           >>> m.x = pyo.Var([1, 2])
           >>> m.eq1 = pyo.Constraint(expr=m.x[1]**2 == 7)
           >>> m.eq2 = pyo.Constraint(expr=m.x[1]*m.x[2] == 3)
           >>> m.eq3 = pyo.Constraint(expr=m.x[1] + 2*m.x[2] == 5)
           >>> igraph = IncidenceGraphInterface(m)
           >>> adj_to_x2 = igraph.get_adjacent_to(m.x[2])
           >>> print([c.name for c in adj_to_x2])
           ['eq2', 'eq3']

        """
        if self._incidence_graph is None:
            raise RuntimeError(
                "Cannot get components adjacent to %s if an incidence graph"
                " is not cached." % component
            )
        _check_unindexed([component])
        M = len(self.constraints)
        N = len(self.variables)
        if component in self._var_index_map:
            vnode = M + self._var_index_map[component]
            adj = self._incidence_graph[vnode]
            adj_comps = [self.constraints[i] for i in adj]
        elif component in self._con_index_map:
            cnode = self._con_index_map[component]
            adj = self._incidence_graph[cnode]
            adj_comps = [self.variables[j - M] for j in adj]
        else:
            raise RuntimeError(
                "Cannot find component %s in the cached incidence graph." % component
            )
        return adj_comps

    def maximum_matching(self, variables=None, constraints=None):
        """Return a maximum cardinality matching of variables and constraints.

        The matching maps constraints to their matched variables.

        Returns
        -------
        ``ComponentMap``
            A map from constraints to their matched variables.

        """
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)
        con_nodes = list(range(len(constraints)))
        matching = maximum_matching(graph, top_nodes=con_nodes)
        # Matching maps constraint nodes to variable nodes. Here we need to
        # know the convention according to which the graph was constructed.
        M = len(constraints)
        return ComponentMap(
            (constraints[i], variables[j - M]) for i, j in matching.items()
        )

    def get_connected_components(self, variables=None, constraints=None):
        """Partition variables and constraints into weakly connected components
        of the incidence graph

        These correspond to diagonal blocks in a block diagonalization of the
        incidence matrix.

        Returns
        -------
        var_blocks: list of lists of variables
            Partition of variables into connected components
        con_blocks: list of lists of constraints
            Partition of constraints into corresponding connected components

        """
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)
        nxc = nx.algorithms.components
        M = len(constraints)
        N = len(variables)
        connected_components = list(nxc.connected_components(graph))

        con_blocks = [
            sorted([i for i in comp if i < M]) for comp in connected_components
        ]
        con_blocks = [[constraints[i] for i in block] for block in con_blocks]
        var_blocks = [
            sorted([j for j in comp if j >= M]) for comp in connected_components
        ]
        var_blocks = [[variables[i - M] for i in block] for block in var_blocks]

        return var_blocks, con_blocks

    # NOTE: That this replaces the <=6.4.4 block_triangularize function
    def map_nodes_to_block_triangular_indices(self, variables=None, constraints=None):
        """Map variables and constraints to indices of their diagonal blocks in
        a block lower triangular permutation

        Returns
        -------
        var_block_map: ``ComponentMap``
            Map from variables to their diagonal blocks in a block
            triangularization
        con_block_map: ``ComponentMap``
            Map from constraints to their diagonal blocks in a block
            triangularization

        """
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)

        M = len(constraints)
        con_nodes = list(range(M))
        sccs = get_scc_of_projection(graph, con_nodes)
        row_idx_map = {r: idx for idx, scc in enumerate(sccs) for r, _ in scc}
        col_idx_map = {c - M: idx for idx, scc in enumerate(sccs) for _, c in scc}
        con_block_map = ComponentMap(
            (constraints[i], idx) for i, idx in row_idx_map.items()
        )
        var_block_map = ComponentMap(
            (variables[j], idx) for j, idx in col_idx_map.items()
        )
        # Switch the order of the maps here to match the method call.
        # Hopefully this does not get too confusing...
        return var_block_map, con_block_map

    def block_triangularize(self, variables=None, constraints=None):
        """Compute an ordered partition of the provided variables and
        constraints such that their incidence matrix is block lower triangular

        Subsets in the partition correspond to the strongly connected components
        of the bipartite incidence graph, projected with respect to a perfect
        matching.

        Returns
        -------
        var_partition: list of lists
            Partition of variables. The inner lists hold unindexed variables.
        con_partition: list of lists
            Partition of constraints. The inner lists hold unindexed constraints.

        Example
        -------

        .. doctest::
           :skipif: not networkx_available

           >>> import pyomo.environ as pyo
           >>> from pyomo.contrib.incidence_analysis import IncidenceGraphInterface
           >>> m = pyo.ConcreteModel()
           >>> m.x = pyo.Var([1, 2])
           >>> m.eq1 = pyo.Constraint(expr=m.x[1]**2 == 7)
           >>> m.eq2 = pyo.Constraint(expr=m.x[1]*m.x[2] == 3)
           >>> igraph = IncidenceGraphInterface(m)
           >>> vblocks, cblocks = igraph.block_triangularize()
           >>> print([[v.name for v in vb] for vb in vblocks])
           [['x[1]'], ['x[2]']]
           >>> print([[c.name for c in cb] for cb in cblocks])
           [['eq1'], ['eq2']]

        .. note::

           **Breaking change in Pyomo 6.5.0**

           The pre-6.5.0 ``block_triangularize`` method returned maps from
           each variable or constraint to the index of its block in a block
           lower triangularization as the original intent of this function
           was to identify when variables do or don't share a diagonal block
           in this partition. Since then, the dominant use case of
           ``block_triangularize`` has been to partition variables and
           constraints into these blocks and inspect or solve each block
           individually. A natural return type for this functionality is the
           ordered partition of variables and constraints, as lists of lists.
           This functionality was previously available via the
           ``get_diagonal_blocks`` method, which was confusing as it did not
           capture that the partition was the diagonal of a block
           *triangularization* (as opposed to diagonalization). The pre-6.5.0
           functionality of ``block_triangularize`` is still available via the
           ``map_nodes_to_block_triangular_indices`` method.

        """
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)

        M = len(constraints)
        con_nodes = list(range(M))
        sccs = get_scc_of_projection(graph, con_nodes)
        var_partition = [[variables[j - M] for _, j in scc] for scc in sccs]
        con_partition = [[constraints[i] for i, _ in scc] for scc in sccs]
        return var_partition, con_partition

    @deprecated(
        msg=(
            "``IncidenceGraphInterface.get_diagonal_blocks`` is deprecated."
            " Please use ``IncidenceGraphInterface.block_triangularize``"
            " instead."
        ),
        version="6.5.0",
    )
    def get_diagonal_blocks(self, variables=None, constraints=None):
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)
        M = len(constraints)
        con_nodes = list(range(M))
        sccs = get_scc_of_projection(graph, con_nodes)
        block_cons = [[constraints[i] for i, _ in scc] for scc in sccs]
        block_vars = [[variables[j - M] for _, j in scc] for scc in sccs]
        return block_vars, block_cons

    def dulmage_mendelsohn(self, variables=None, constraints=None):
        """Partition variables and constraints according to the Dulmage-
        Mendelsohn characterization of the incidence graph

        Variables are partitioned into the following subsets:

        - **unmatched** - Variables not matched in a particular maximum
          cardinality matching
        - **underconstrained** - Variables that *could possibly be* unmatched
          in a maximum cardinality matching
        - **square** - Variables in the well-constrained subsystem
        - **overconstrained** - Variables matched with constraints that can
          possibly be unmatched

        Constraints are partitioned into the following subsets:

        - **underconstrained** - Constraints matched with variables that can
          possibly be unmatched
        - **square** - Constraints in the well-constrained subsystem
        - **overconstrained** - Constraints that *can possibly be* unmatched
          with a maximum cardinality matching
        - **unmatched** - Constraints that were not matched in a particular
          maximum cardinality matching

        While the Dulmage-Mendelsohn decomposition does not specify an order
        within any of these subsets, the order returned by this function
        preserves the maximum matching that is used to compute the decomposition.
        That is, zipping "corresponding" variable and constraint subsets yields
        pairs in this maximum matching. For example:

        .. doctest::
           :hide:
           :skipif: not (networkx_available and scipy_available)

           >>> # Hidden code block creating a dummy model so the following doctest runs
           >>> import pyomo.environ as pyo
           >>> from pyomo.contrib.incidence_analysis import IncidenceGraphInterface
           >>> model = pyo.ConcreteModel()
           >>> model.x = pyo.Var([1,2,3])
           >>> model.eq = pyo.Constraint(expr=sum(m.x[:]) == 1)

        .. doctest::
           :skipif: not (networkx_available and scipy_available)

           >>> igraph = IncidenceGraphInterface(model)
           >>> var_dmpartition, con_dmpartition = igraph.dulmage_mendelsohn()
           >>> vdmp = var_dmpartition
           >>> cdmp = con_dmpartition
           >>> matching = list(zip(
           ...     vdmp.underconstrained + vdmp.square + vdmp.overconstrained,
           ...     cdmp.underconstrained + cdmp.square + cdmp.overconstrained,
           ... ))
           >>> # matching is a valid maximum matching of variables and constraints!

        Returns
        -------
        var_partition: ``ColPartition`` named tuple
            Partitions variables into square, underconstrained, overconstrained,
            and unmatched.
        con_partition: ``RowPartition`` named tuple
            Partitions constraints into square, underconstrained,
            overconstrained, and unmatched.

        Example
        -------

        .. doctest::
           :skipif: not networkx_available

           >>> import pyomo.environ as pyo
           >>> from pyomo.contrib.incidence_analysis import IncidenceGraphInterface
           >>> m = pyo.ConcreteModel()
           >>> m.x = pyo.Var([1, 2])
           >>> m.eq1 = pyo.Constraint(expr=m.x[1]**2 == 7)
           >>> m.eq2 = pyo.Constraint(expr=m.x[1]*m.x[2] == 3)
           >>> m.eq3 = pyo.Constraint(expr=m.x[1] + 2*m.x[2] == 5)
           >>> igraph = IncidenceGraphInterface(m)
           >>> var_dmp, con_dmp = igraph.dulmage_mendelsohn()
           >>> print([v.name for v in var_dmp.overconstrained])
           ['x[1]', 'x[2]']
           >>> print([c.name for c in con_dmp.overconstrained])
           ['eq1', 'eq2']
           >>> print([c.name for c in con_dmp.unmatched])
           ['eq3']

        """
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)
        M = len(constraints)
        top_nodes = list(range(M))
        row_partition, col_partition = dulmage_mendelsohn(graph, top_nodes=top_nodes)
        con_partition = RowPartition(
            *[[constraints[i] for i in subset] for subset in row_partition]
        )
        var_partition = ColPartition(
            *[[variables[i - M] for i in subset] for subset in col_partition]
        )
        # Switch the order of the maps here to match the method call.
        # Hopefully this does not get too confusing...
        return var_partition, con_partition

    def remove_nodes(self, variables=None, constraints=None):
        """Removes the specified variables and constraints (columns and
        rows) from the cached incidence matrix.

        This is a "projection" of the variable and constraint vectors, rather
        than something like a vertex elimination. For the puropse of this
        method, there is no need to distinguish between variables and
        constraints. However, we provide the "constraints" argument so a call
        signature similar to other methods in this class is still valid.

        Parameters
        ----------
        variables: list
            VarData objects whose nodes will be removed from the incidence graph
        constraints: list
            ConData objects whose nodes will be removed from the incidence graph

        .. note::

           **Deprecation in Pyomo v6.7.2**

           The pre-6.7.2 implementation of ``remove_nodes`` allowed variables and
           constraints to remove to be specified in a single list. This made
           error checking difficult, and indeed, if invalid components were
           provided, we carried on silently instead of throwing an error or
           warning. As part of a fix to raise an error if an invalid component
           (one that is not part of the incidence graph) is provided, we now require
           variables and constraints to be specified separately.

        """
        if constraints is None:
            constraints = []
        if variables is None:
            variables = []
        if self._incidence_graph is None:
            raise RuntimeError(
                "Attempting to remove variables and constraints from cached "
                "incidence matrix,\nbut no incidence matrix has been cached."
            )

        vars_to_validate = []
        cons_to_validate = []
        depr_msg = (
            "In IncidenceGraphInterface.remove_nodes, passing variables and"
            " constraints in the same list is deprecated. Please separate your"
            " variables and constraints and pass them in the order variables,"
            " constraints."
        )
        if any(var in self._con_index_map for var in variables) or any(
            con in self._var_index_map for con in constraints
        ):
            deprecation_warning(depr_msg, version="6.7.2")
        # If we received variables/constraints in the same list, sort them.
        # Any unrecognized objects will be caught by _validate_input.
        for var in variables:
            if var in self._con_index_map:
                cons_to_validate.append(var)
            else:
                vars_to_validate.append(var)
        for con in constraints:
            if con in self._var_index_map:
                vars_to_validate.append(con)
            else:
                cons_to_validate.append(con)

        variables, constraints = self._validate_input(
            vars_to_validate, cons_to_validate
        )
        v_exclude = ComponentSet(variables)
        c_exclude = ComponentSet(constraints)
        vars_to_include = [v for v in self.variables if v not in v_exclude]
        cons_to_include = [c for c in self.constraints if c not in c_exclude]
        incidence_graph = self._extract_subgraph(vars_to_include, cons_to_include)
        # update attributes
        self._variables = vars_to_include
        self._constraints = cons_to_include
        self._incidence_graph = incidence_graph
        self._var_index_map = ComponentMap(
            (var, i) for i, var in enumerate(vars_to_include)
        )
        self._con_index_map = ComponentMap(
            (con, i) for i, con in enumerate(cons_to_include)
        )

    def plot(self, variables=None, constraints=None, title=None, show=True):
        """Plot the bipartite incidence graph of variables and constraints"""
        variables, constraints = self._validate_input(variables, constraints)
        graph = self._extract_subgraph(variables, constraints)
        M = len(constraints)

        left_nodes = list(range(M))
        pos_dict = nx.drawing.bipartite_layout(graph, nodes=left_nodes)

        edge_x = []
        edge_y = []
        for start_node, end_node in graph.edges():
            x0, y0 = pos_dict[start_node]
            x1, y1 = pos_dict[end_node]
            edge_x.append(x0)
            edge_x.append(x1)
            edge_x.append(None)
            edge_y.append(y0)
            edge_y.append(y1)
            edge_y.append(None)
        edge_trace = plotly.graph_objects.Scatter(
            x=edge_x,
            y=edge_y,
            line=dict(width=0.5, color="#888"),
            hoverinfo="none",
            mode="lines",
        )

        node_x = []
        node_y = []
        node_text = []
        node_color = []
        for node in graph.nodes():
            x, y = pos_dict[node]
            node_x.append(x)
            node_y.append(y)
            if node < M:
                # According to convention, we are a constraint node
                c = constraints[node]
                node_color.append("red")
                body_text = "<br>".join(
                    textwrap.wrap(str(c.body), width=120, subsequent_indent="    ")
                )
                node_text.append(
                    f"{str(c)}<br>lb: {str(c.lower)}<br>body: {body_text}<br>"
                    f"ub: {str(c.upper)}<br>active: {str(c.active)}"
                )
            else:
                # According to convention, we are a variable node
                v = variables[node - M]
                node_color.append("blue")
                node_text.append(
                    f"{str(v)}<br>lb: {str(v.lb)}<br>ub: {str(v.ub)}<br>"
                    f"value: {str(v.value)}<br>domain: {str(v.domain)}<br>"
                    f"fixed: {str(v.is_fixed())}"
                )
        node_trace = plotly.graph_objects.Scatter(
            x=node_x,
            y=node_y,
            mode="markers",
            hoverinfo="text",
            text=node_text,
            marker=dict(color=node_color, size=10),
        )
        fig = plotly.graph_objects.Figure(data=[edge_trace, node_trace])
        if title is not None:
            fig.update_layout(title=dict(text=title))
        if show:
            fig.show()

    def add_edge(self, variable, constraint):
        """Adds an edge between variable and constraint in the incidence graph

        Parameters
        ----------
        variable: VarData
            A variable in the graph
        constraint: ConstraintData
            A constraint in the graph
        """
        if self._incidence_graph is None:
            raise RuntimeError(
                "Attempting to add edge in an incidence graph from cached "
                "incidence graph,\nbut no incidence graph has been cached."
            )

        if variable not in self._var_index_map:
            raise RuntimeError("%s is not a variable in the incidence graph" % variable)

        if constraint not in self._con_index_map:
            raise RuntimeError(
                "%s is not a constraint in the incidence graph" % constraint
            )

        var_id = self._var_index_map[variable] + len(self._con_index_map)
        con_id = self._con_index_map[constraint]

        self._incidence_graph.add_edge(var_id, con_id)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# The log should be imported first so that the Pyomo LogHandler can be
# set up as soon as possible
from . import log
from . import envvar

from .factory import Factory

from .fileutils import (
    Executable,
    Library,
    # The following will be deprecated soon
    register_executable,
    registered_executable,
    unregister_executable,
)
from . import config, dependencies, shutdown, timing
from .deprecation import deprecated
from .errors import DeveloperError
from ._command import pyomo_command, get_pyomo_commands

#
# declare deprecation paths for removed modules
#
from .deprecation import moved_module

moved_module('pyomo.common.getGSL', 'pyomo.common.gsl', version='6.5.0')
moved_module('pyomo.common.plugin', 'pyomo.common.plugin_base', version='6.5.0')
del moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: UTF-8 -*-
"""Tests disjunct reclassifier transformation."""

import pyomo.common.unittest as unittest
from pyomo.core import (
    Block,
    ConcreteModel,
    TransformationFactory,
    RangeSet,
    Constraint,
    Var,
)
from pyomo.gdp import Disjunct, Disjunction, GDP_Error


class TestDisjunctReclassify(unittest.TestCase):
    """Tests reclassification of disjuncts."""

    def test_deactivated_parent_disjunct(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d1.sub1 = Disjunct()
        m.d1.sub2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.sub1, m.d1.sub2])
        m.d1.deactivate()
        TransformationFactory('gdp.reclassify').apply_to(m)
        self.assertIs(m.d1.ctype, Block)
        self.assertIs(m.d1.sub1.ctype, Block)
        self.assertIs(m.d1.sub2.ctype, Block)

    def test_deactivated_parent_block(self):
        m = ConcreteModel()
        m.d1 = Block()
        m.d1.sub1 = Disjunct()
        m.d1.sub2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.sub1, m.d1.sub2])
        m.d1.deactivate()
        TransformationFactory('gdp.reclassify').apply_to(m)
        self.assertIs(m.d1.ctype, Block)
        self.assertIs(m.d1.sub1.ctype, Block)
        self.assertIs(m.d1.sub2.ctype, Block)

    def test_active_parent_disjunct(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d1.sub1 = Disjunct()
        m.d1.sub2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.sub1, m.d1.sub2])
        with self.assertRaises(GDP_Error):
            TransformationFactory('gdp.reclassify').apply_to(m)

    def test_active_parent_disjunct_target(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d1.sub1 = Disjunct()
        m.d1.sub2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.sub1, m.d1.sub2])
        TransformationFactory('gdp.bigm').apply_to(m, targets=m.d1.disj)
        m.d1.indicator_var.fix(1)
        TransformationFactory('gdp.reclassify').apply_to(m)
        self.assertIs(m.d1.ctype, Block)
        self.assertIs(m.d1.sub1.ctype, Block)
        self.assertIs(m.d1.sub2.ctype, Block)

    def test_active_parent_block(self):
        m = ConcreteModel()
        m.d1 = Block()
        m.d1.sub1 = Disjunct()
        m.d1.sub2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.sub1, m.d1.sub2])
        with self.assertRaises(GDP_Error):
            TransformationFactory('gdp.reclassify').apply_to(m)

    def test_deactivate_nested_disjunction(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d1.d1 = Disjunct()
        m.d1.d2 = Disjunct()
        m.d1.disj = Disjunction(expr=[m.d1.d1, m.d1.d2])
        m.d2 = Disjunct()
        m.disj = Disjunction(expr=[m.d1, m.d2])
        m.d1.deactivate()
        TransformationFactory('gdp.bigm').apply_to(m)
        # for disj in m.component_data_objects(Disjunction, active=True):
        #     print(disj.name)
        # There should be no active Disjunction objects.
        self.assertIsNone(
            next(m.component_data_objects(Disjunction, active=True), None)
        )

    def test_do_not_reactivate_disjuncts_with_abandon(self):
        m = ConcreteModel()
        m.x = Var()
        m.s = RangeSet(4)
        m.d = Disjunct(m.s)
        m.d[2].bad_constraint_should_not_be_active = Constraint(expr=m.x >= 1)
        m.disj1 = Disjunction(expr=[m.d[1], m.d[2]])
        m.disj2 = Disjunction(expr=[m.d[3], m.d[4]])
        m.d[1].indicator_var.fix(1)
        m.d[2].deactivate()
        TransformationFactory('gdp.bigm').apply_to(m)
        self.assertFalse(m.d[2].active)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.contrib.pynumero.dependencies import numpy_available, scipy_available

if not numpy_available or not scipy_available:
    raise unittest.SkipTest("pynumero linear solver tests require numpy and scipy")
from pyomo.contrib.pynumero.sparse import BlockVector, BlockMatrix
from scipy.sparse import coo_matrix, spmatrix
import numpy as np
from pyomo.contrib.pynumero.linalg.base import LinearSolverInterface, LinearSolverStatus
from pyomo.contrib.pynumero.linalg.ma27_interface import MA27
from pyomo.contrib.pynumero.linalg.ma57_interface import MA57
from pyomo.contrib.pynumero.linalg.ma27 import MA27Interface
from pyomo.contrib.pynumero.linalg.ma57 import MA57Interface
from pyomo.contrib.pynumero.linalg.scipy_interface import ScipyLU, ScipyIterative
from scipy.sparse.linalg import gmres
from pyomo.contrib.pynumero.linalg.mumps_interface import (
    mumps_available,
    MumpsCentralizedAssembledLinearSolver,
)


class TestLinearSolvers(unittest.TestCase):
    def create_blocks(self, m: np.ndarray, x: np.ndarray):
        m = coo_matrix(m)
        r = m * x
        bm = BlockMatrix(2, 2)
        bm.set_block(0, 0, m.copy())
        bm.set_block(1, 1, m.copy())
        br = BlockVector(2)
        br.set_block(0, r.copy())
        br.set_block(1, r.copy())
        bx = BlockVector(2)
        bx.set_block(0, x.copy())
        bx.set_block(1, x.copy())
        return bm, bx, br

    def solve_helper(self, m: np.ndarray, x: np.ndarray, solver: LinearSolverInterface):
        bm, bx, br = self.create_blocks(m, x)
        bx2, res = solver.solve(bm, br)
        self.assertEqual(res.status, LinearSolverStatus.successful)
        err = np.max(np.abs(bx - bx2))
        self.assertAlmostEqual(err, 0)

    def symmetric_helper(self, solver: LinearSolverInterface):
        m = np.array([[1, 2], [2, -1]], dtype=np.double)
        x = np.array([4, 7], dtype=np.double)
        self.solve_helper(m, x, solver)

    def unsymmetric_helper(self, solver: LinearSolverInterface):
        m = np.array([[1, 2], [0, -1]], dtype=np.double)
        x = np.array([4, 7], dtype=np.double)
        self.solve_helper(m, x, solver)

    def singular_helper(self, solver: LinearSolverInterface):
        m = np.array([[1, 1], [1, 1]], dtype=np.double)
        x = np.array([4, 7], dtype=np.double)
        bm, bx, br = self.create_blocks(m, x)
        br.get_block(0)[1] += 1
        br.get_block(1)[1] += 1
        bx2, res = solver.solve(bm, br, raise_on_error=False)
        self.assertNotEqual(res.status, LinearSolverStatus.successful)

    @unittest.skipIf(not MA27Interface.available(), reason="MA27 not available")
    def test_ma27(self):
        solver = MA27()
        self.symmetric_helper(solver)
        self.singular_helper(solver)

    @unittest.skipIf(not MA57Interface.available(), reason="MA57 not available")
    def test_ma57(self):
        solver = MA57()
        self.symmetric_helper(solver)
        self.singular_helper(solver)

    def test_scipy_direct(self):
        solver = ScipyLU()
        self.symmetric_helper(solver)
        self.unsymmetric_helper(solver)
        self.singular_helper(solver)

    def test_scipy_iterative(self):
        solver = ScipyIterative(gmres)
        solver.options["atol"] = 1e-8
        self.symmetric_helper(solver)
        self.unsymmetric_helper(solver)
        self.singular_helper(solver)

    @unittest.skipIf(not mumps_available, reason="mumps not available")
    def test_mumps(self):
        solver = MumpsCentralizedAssembledLinearSolver(sym=2)
        self.symmetric_helper(solver)
        self.singular_helper(solver)
        solver = MumpsCentralizedAssembledLinearSolver(sym=0)
        self.unsymmetric_helper(solver)
